/*	Copyright  (c)	Günter Woigk 2009 - 2015
					mailto:kio@little-bat.de

	This file is free software

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	• Redistributions of source code must retain the above copyright notice,
	  this list of conditions and the following disclaimer.
	• Redistributions in binary form must reproduce the above copyright notice,
	  this list of conditions and the following disclaimer in the documentation
	  and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
	THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
	PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


	==============================================================================


	Microcode for the K1-16/16 CPU

	2013-03-07
	2015-12-22


The microcode implements 3 different code models:

microcode:	located in the microcode rom and executed by the control unit hardware.
			microcodes are addressed with the microcode address counter.
		  	subroutines are called with 'jsr' (actually 5 microcodes in total!)
			and return with 'ret' which actually consists of 2 microcodes, jp and a nop,
			but one last microcode can sometimes be postponed into the 'nop'.

		  •	labels are all_lowercase.
		  •	arguments and return values are typically passed in registers.
		  • subroutines try to preserve registers.

Millicode:	located in the microcode rom and addressed with the ip register.
			millicodes are implemented in microcode.
			but may itself switch to millicode with the 'millicoded' microcode.
		  	millicodes are the implementation's address with condition 'bit15'.
			millicodes jump to the next millicode with the 'next_mc' microcode,
			or the 'Ret' millicode, if they are millicoded.
			note: 'millicoded' pushes the current ip and 'Ret' restores it.

		  •	labels (millicodes!) are CamelCased.
		  •	arguments are passed on the data stack with the top value in the alu register.
		  • most (simple) millicodes preserve all common registers.
		  •	subroutines are implemented as new millicodes. (there is no 'Jsr')

OPCODE:		located in ram and addressed with the ip register.
			opcodes are implemented in microcode.
		  	opcodes are the implementation's address with
				- bit15 = code plane of the address and
				- bit14 = 1 to not clear the low byte or
				- bit14 = 0 to clear the low byte which may contain one byte of data (cmd_lo).
			opcodes jump to the next opcode with the 'next' microcode,
			which actually consists of 2 microcodes, the jp and a nop,
			but one last microcode can frequently be postponed into the 'nop'.

		  •	labels (opcodes!) are ALL_UPPERCASE.
		  •	arguments are passed on the data stack with the top value in the alu register.
		  • most (simple) opcodes preserve all common registers.
		  •	subroutines (in ram) are implemented with opcodes.
			they are called with 'JSR' and return with 'RET' opcode.
		  • reading of the next opcode can be superseded with the interrupt vector.
			due to the hardware this results in a jump to the read opcode address & $F000.
			address $0000 is the reset entry, therefore opcodes must not start in range $0000-$0FFF.
*/


#include	"options.h"				// options for cc and asm
#include	"asm/defs.asm"			// symbolic names for registers and so on
#include	"asm/globals.asm"		// global data
#include	"asm/macros.asm"		// instruction macros
#include 	"asm/microcodes.asm"	// #defines for microcodes implemented  in c


#include	"asm/exception.asm"
#include	"asm/ramtest.asm"
#include	"asm/mem.asm"
#include	"asm/ldir.asm"
#include	"asm/interrupt.asm"
#include	"asm/millicode.asm"
#include	"asm/millicode_int.asm"
#include	"asm/millicode_float.asm"
#include	"asm/millicode_str.asm"
#include	"asm/combicodes.asm"
#include	"asm/opcodes.asm"


#include	"stdc.cc"
#include	"long.cc"
#include	"str.cc"
#include	"sort.cc"

#if include_sio_driver_in_rom
#include	"sio.cc"
#endif

#if include_floating_point
#include	"float.cc"
#endif


// ______________________________________________________
// "Production" Microcode:

#if !include_tests
#include	"i2c.cc"
#include	"dev.cc"
#include	"FD.cc"
#include	"SerialDevice.cc"
#include	"Device.cc"
#include	"BlockDevice.cc"
#if include_filesystem
#include	"Block.cc"
#include	"File.cc"
#include	"Dir.cc"
#endif
#include	"99_bottles_of_beer.cc"
#include	"bcc.cc"
#include	"shell.cc"
#endif


// ______________________________________________________
// "Test" Microcode:

#if include_tests
//#include	"test/test_helper.cc"

#if test_i2c
#include	"test/Block.cc"
#include	"test/Device.cc"
#include	"test/BlockDevice.cc"
#include	"i2c.cc"
#include	"test/test_i2c.cc"
#endif

#if test_millicode
#include	"test_millicode.asm"
#endif

#if test_long
#include	"test/test_long.cc"
#endif

#if test_float
#include	"test/test_float.cc"
#endif

#if test_misc
#include	"test/test_misc.cc"
#endif

#endif




// ======================================================
//	RESET
// ======================================================

//	Reset entry condition:
//	• $7FFF with 'cmddis=1' and 'cond=0' (for next instr) forced by INIT
//	• $0000 is executed
//
: reset_7fff = $3FFF:1
		clr	reset			// 'nop' ((cmd_dis version))

: reset = $0000
		clr halt			// don't halt cpu if no interrupt pending
		di					// disable interrupt processing ((in "ld cmd,(ip++)" aka "next"))

		set	clock4			// set full speed
		set	clock2			// ((except if clock_predivider > 2))



// ======================================
// 	CPU Test
// ======================================

		clr led_red			// see "exceptions.asm" for LED codes
		set led_yel
		set led_grn			// -/yel/grn	--> booting, cpu self test

		ld	alu,1
		do
			ld	d0,alu
			ld	a0,d0
			ld	d1,a0
			ld	a1,d1
			ld	d2,a1
			ld	atmp,d2
			ld	dtmp,atmp
			ld	hp,dtmp
			ld	sl,hp,cy
			ld	sp,sl	: bit15
			if 0
				ld	sr,sp,nc
			else
				ld	sr,sp,cy
			then
			ld	ip,sr
			ld	swap,ip
			ld	swap,swap
			ld	d2ar,swap

			equ	d2ar
			TRAP !z

			add	alu				// nodelay
		until z



// ======================================
// 	Test and Clear Ram
// ======================================

		Millicode
		mc	ClrLedGrn			// -/yel/- 	--> booting, ram test
		mc	Ramtest				// 			--> alu = 0


// ======================================
// 	Init Global Data
// ======================================

		mc	ClrLedYel			// -/-/grn 	--> booting, initializing
		mc	SetLedGrn

// Initialize non-zero global data:

// interrupt vectors:

		mc	Int8+15, PopHP		// int_vector[16]	all 0x0000			note: hp points to last byte, empty heap = $ffff
		mc	Ival $ffff			// int_mask 		of installed devices

// memory management:

		mc	Ival mem_globs_end	// gdata_ptr
		mc	Ival $100			// gcode_ptr 		= start of space for new dicts

		mc	Ival $4000			// hp_base 			= mem_size/4
		mc	Ival $8000			// sp_base 			= mem_size/2

		mc	Dup					// mem_data_start 	:= sp_base
		mc	Dup					// mem_data_end	  	:= sp_base

		mc	Ival $FFFE			// mem_ptr_start	start of mem pointers = mem_end-2
		mc	Dup					// mem_free_ptr		start of linked list of free mem pointers = mem_end-2
		mc	Int8+0				// mem_end			end of physical ram	= mem_end: $0000	(TODO: eliminate?)

// misc:
		mc	Ival Ret			// opcode Ret: return to millicode

#if include_filesystem
		mc	Istr ".trash"		// cstr ".trash" = trash file name
#endif

#if include_sio_driver_in_rom
		mc	Ival SIO_GETCTL_PROC
		mc	Ival SIO_SETCTL_PROC
		mc	Ival SIO_GETC_PROC
		mc	Ival SIO_GETS_PROC
		mc	Ival SIO_PUTC_PROC
		mc	Ival SIO_PUTS_PROC
		mc	Ival SIO_IRPT_PROC
#endif

// Initialize registers:
// must be done before calling millicoded words

		mc	Gget+hp_base,PopHP	// reset hp		note: also writes TOP value to ram
		mc	Gget+sp_base,PopSP	// reset sp

#if include_floating_point
		mc	init_float			// init globals in float.cc
#endif

		mc	ClrLedGrn			// -/-/-		"RUNNING"
		mc	STOP2


// =======================================================
// 	Test Code
// =======================================================


#if include_tests
#if test_millicode
		mc	TestMillicode
#endif

#if test_misc
		mc	TestMisc
#endif

#if test_i2c
		mc	TestI2C
#endif

#if test_long
		mc TestLong
#endif

#if test_float
		mc TestFloat
#endif

		mc	Istr "---OK!---", LogStr
		mc	SetLedYel
		mc	SetLedGrn
		mc	STOP2
#endif



// =======================================================
// 	Load and Init Device Drivers, Start Shell
// =======================================================


#if !include_tests

		mc	init_devices		// init k1-bus devices in dev.cc
		mc	ClrLedGrn			// -/-/-	--> running: all Leds off

		mc STOP2
		mc	run_shell
		mc	reset
#endif













